#include "File.h"
#include "OAL.h"

#if defined(HAVE_DIRENT_H)
#include <dirent.h>		// struct stat
#endif

namespace AScript {

//-----------------------------------------------------------------------------
// Console
//-----------------------------------------------------------------------------
Console::Console() : _pCodecRead(NULL), _pCodecWrite(NULL)
{
}

Console::~Console()
{
	ReleaseCodec();
}

bool Console::InstallCodec(const char *encoding, bool processEOLFlag)
{
	if (encoding == NULL) encoding = "us-ascii";
	CodecFactory *pCodecFactory = CodecFactory::Lookup(encoding);
	if (pCodecFactory == NULL) return false;
	ReleaseCodec();
	_encoding = encoding;
	_pCodecRead = pCodecFactory->CreateDecoder(processEOLFlag);
	_pCodecWrite = pCodecFactory->CreateEncoder(processEOLFlag);
	return true;
}

void Console::ReleaseCodec()
{
	_encoding.clear();
	delete _pCodecRead;
	delete _pCodecWrite;
	_pCodecRead = NULL;
	_pCodecWrite = NULL;
}

void Console::PutChar(char ch)
{
	if (_pCodecWrite == NULL) {
		PutRawChar(ch);
	} else {
		char chConv;
		Codec::Result rtn = _pCodecWrite->FeedChar(ch, chConv);
		if (rtn == Codec::RESULT_Complete) {
			PutRawChar(chConv);
			while (_pCodecWrite->FollowChar(chConv)) PutRawChar(chConv);
		} else if (rtn == Codec::RESULT_Error) {
			// nothing to do
		}
	}
}

int Console::GetChar(Error *pError)
{
	if (pError != NULL) *pError = ERROR_None;
	if (_pCodecRead == NULL) return GetRawChar();
	char chConv;
	if (_pCodecRead->FollowChar(chConv)) return static_cast<unsigned char>(chConv);
	for (;;) {
		int ch = GetRawChar();
		if (ch < 0) return ch;
		Codec::Result rtn = _pCodecRead->FeedChar(static_cast<char>(ch), chConv);
		if (rtn == Codec::RESULT_Complete) {
			break;
		} else if (rtn == Codec::RESULT_Error) {
			if (pError != NULL) *pError = ERROR_Codec;
		}
	}
	return static_cast<unsigned char>(chConv);
}

void Console::Print(const char *str)
{
	for ( ; *str != '\0'; str++) PutChar(*str);
	FlushConsole();
}

void Console::Println(const char *str)
{
	Print(str);
	PutChar('\n');
	FlushConsole();
}

void Console::Printf(Signal sig, const char *format, const ValueList &valList)
{
	String str = Formatter::Format(sig, format, valList);
	if (sig.IsSignalled()) return;
	Print(str.c_str());
}

void Console::PrintSignal(const Signal &sig)
{
	if (sig.GetSignalType() == SIGTYPE_Error) {
		Println(sig.GetErrString().c_str());
	} else {
		Println(sig.GetSignalName());
	}
}

int Console::GetRawChar()
{
	return -1;
}

void Console::PutRawChar(char ch)
{
}

void Console::FlushConsole()
{
}

//-----------------------------------------------------------------------------
// FileStat
//-----------------------------------------------------------------------------
#if defined(HAVE_WINDOWS_H)
FileStat::FileStat(const char *pathName, const BY_HANDLE_FILE_INFORMATION &attrData) :
	_pathName(pathName), _attr(0), _bytes(attrData.nFileSizeLow),
	_atime(attrData.ftLastAccessTime, true),
	_mtime(attrData.ftLastWriteTime, true),
	_ctime(attrData.ftCreationTime, true), _uid(0), _gid(0)
{
	if (attrData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
		_attr |= ATTR_Dir;
	} else {
		_attr |= ATTR_Reg;
	}
	if (attrData.dwFileAttributes & FILE_ATTRIBUTE_READONLY) {
		_attr |= 0666;
	} else {
		_attr |= 0777;
	}
}

FileStat::FileStat(const char *pathName, const WIN32_FILE_ATTRIBUTE_DATA &attrData) :
	_pathName(pathName), _attr(0), _bytes(attrData.nFileSizeLow),
	_atime(attrData.ftLastAccessTime, true),
	_mtime(attrData.ftLastWriteTime, true),
	_ctime(attrData.ftCreationTime, true), _uid(0), _gid(0)
{
	if (attrData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
		_attr |= ATTR_Dir;
	} else {
		_attr |= ATTR_Reg;
	}
	if (attrData.dwFileAttributes & FILE_ATTRIBUTE_READONLY) {
		_attr |= 0666;
	} else {
		_attr |= 0777;
	}
}
#else
FileStat::FileStat(const char *pathName, const struct stat &stat) :
	_pathName(pathName), _attr(0), _bytes(stat.st_size),
	_atime(stat.st_atime, true),
	_mtime(stat.st_mtime, true),
	_ctime(stat.st_ctime, true), _uid(stat.st_uid), _gid(stat.st_gid)
{
	if (S_ISDIR(stat.st_mode)) _attr |= ATTR_Dir;
	if (S_ISCHR(stat.st_mode)) _attr |= ATTR_Chr;
	if (S_ISBLK(stat.st_mode)) _attr |= ATTR_Blk;
	if (S_ISREG(stat.st_mode)) _attr |= ATTR_Reg;
	if (S_ISFIFO(stat.st_mode)) _attr |= ATTR_Fifo;
	if (S_ISLNK(stat.st_mode)) _attr |= ATTR_Lnk;
	if (S_ISSOCK(stat.st_mode)) _attr |= ATTR_Sock;
	_attr |= (stat.st_mode & 0777);
}
#endif

//-----------------------------------------------------------------------------
// FileLister
//-----------------------------------------------------------------------------
FileLister::~FileLister()
{
}

#if defined(HAVE_WINDOWS_H)
//-----------------------------------------------------------------------------
// FileLister_Win32
//-----------------------------------------------------------------------------
class FileLister_Win32 : public FileLister {
private:
	HANDLE _hFind;
public:
	inline FileLister_Win32(const char *dirName) :
						FileLister(dirName), _hFind(INVALID_HANDLE_VALUE) {}
	virtual ~FileLister_Win32();
	virtual bool Next(FileEntry &entry);
};

FileLister *FileLister::CreateInstance(const char *dirName)
{
	return new FileLister_Win32(dirName);
}

FileLister_Win32::~FileLister_Win32()
{
	::FindClose(_hFind);
}

bool FileLister_Win32::Next(FileEntry &entry)
{
	WIN32_FIND_DATA findData;
	if (_hFind == INVALID_HANDLE_VALUE) {
		String pathName(_dirName);
		if (!pathName.empty()) pathName += File::Separator;
		pathName += "*.*";
		_hFind = ::FindFirstFile(pathName.c_str(), &findData);
		if (_hFind == INVALID_HANDLE_VALUE) return false;
	} else if (!::FindNextFile(_hFind, &findData)) {
		return false;
	}
	while (::strcmp(findData.cFileName, ".") == 0 ||
			::strcmp(findData.cFileName, "..") == 0) {
		if (!::FindNextFile(_hFind, &findData)) return false;
	}
	entry = FileEntry(findData.cFileName,
		(findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)? true : false);
	return true;
}

#else

//-----------------------------------------------------------------------------
// FileLister_Linux
//-----------------------------------------------------------------------------
class FileLister_Linux : public FileLister {
private:
	DIR *_pDir;
public:
	FileLister_Linux(const char *dirName);
	virtual ~FileLister_Linux();
	virtual bool Next(FileEntry &entry);
};

FileLister *FileLister::CreateInstance(const char *dirName)
{
	return new FileLister_Linux(dirName);
}

FileLister_Linux::FileLister_Linux(const char *dirName) :
									FileLister(dirName), _pDir(NULL)
{
}

FileLister_Linux::~FileLister_Linux()
{
	if (_pDir != NULL) {
		closedir(_pDir);
	}
}

bool FileLister_Linux::Next(FileEntry &entry)
{
	if (_pDir == NULL) {
		_pDir = opendir(_dirName.empty()? "." : _dirName.c_str());
		if (_pDir == NULL) return false;
	}
	struct dirent *pEnt = readdir(_pDir);
	if (pEnt == NULL) return false;
	entry = FileEntry(pEnt->d_name, (pEnt->d_type == DT_DIR));
	return true;
}

#endif

//-----------------------------------------------------------------------------
// File
//-----------------------------------------------------------------------------
const char File::SeparatorWin = '\\';
const char File::SeparatorUnix = '/';
String *File::_pWorkingDir = NULL;

File::~File()
{
	Close();
}

#if defined(HAVE_WINDOWS_H)
const char File::Separator = SeparatorWin;

File::File() : _hFile(INVALID_HANDLE_VALUE), _needCloseFlag(false), _accessMode(ACCESS_Read)
{
}

bool File::Open(Signal sig, const char *fileName,
								const char *mode, const char *encoding)
{
	Close();
	bool processEOLFlag = false;
	do {
		const char *p = mode;
		if (*p == 'r') {
			_accessMode = ACCESS_Read;
		} else if (*p == 'w') {
			_accessMode = ACCESS_Write;
		} else if (*p == 'a') {
			_accessMode = ACCESS_Write;
		} else {
			sig.SetError(ERR_IOError, "invalid open mode");
			return false;
		}
		p++;
		for ( ; *p != '\0'; p++) {
			char ch = *p;
			if (ch == '+') {
				_accessMode = ACCESS_ReadWrite;
			} else if (ch == 't') {
				processEOLFlag = true;
			} else if (ch == 'b') {
				processEOLFlag = false;
			} else {
				sig.SetError(ERR_IOError, "invalid open mode");
				return false;
			}
		}
	} while (0);
	if (!InstallCodec(encoding, processEOLFlag)) {
		sig.SetError(ERR_CodecError, "unsupported encoding '%s'", encoding);
		return false;
	}
	_fileName = MakeAbsPath(fileName);
	DWORD dwDesiredAccess =
			(_accessMode == ACCESS_Read)? GENERIC_READ :
			(_accessMode == ACCESS_Write)? GENERIC_WRITE :
			(_accessMode == ACCESS_ReadWrite)? GENERIC_READ | GENERIC_WRITE :
			GENERIC_READ;
	DWORD dwShareMode = FILE_SHARE_READ;
	DWORD dwCreationDisposition =
			(_accessMode == ACCESS_Read)? OPEN_EXISTING :
			(_accessMode == ACCESS_Write)? CREATE_ALWAYS :
			(_accessMode == ACCESS_ReadWrite)? OPEN_ALWAYS :
			OPEN_EXISTING;
	_hFile = ::CreateFile(_fileName.c_str(), dwDesiredAccess, dwShareMode,
					NULL, dwCreationDisposition, FILE_ATTRIBUTE_NORMAL, NULL);
	if (_hFile == INVALID_HANDLE_VALUE) {
		sig.SetError(ERR_IOError, "can't open file '%s'", ExtractBaseName(fileName).c_str());
		return false;
	}
	_needCloseFlag = true;
	return true;
}

bool File::OpenStdin()
{
	Close();
	_hFile = ::GetStdHandle(STD_INPUT_HANDLE);
	if (_hFile != INVALID_HANDLE_VALUE) {
		_fileName = "stdin", _accessMode = ACCESS_Read;
		InstallCodec("cp932", true);
	}
	return true;
}

bool File::OpenStdout()
{
	Close();
	_hFile = ::GetStdHandle(STD_OUTPUT_HANDLE);
	if (_hFile != INVALID_HANDLE_VALUE) {
		_fileName = "stdout", _accessMode = ACCESS_Write;
		InstallCodec("cp932", true);
	}
	return true;
}

bool File::OpenStderr()
{
	Close();
	_hFile = ::GetStdHandle(STD_ERROR_HANDLE);
	if (_hFile != INVALID_HANDLE_VALUE) {
		_fileName = "stderr", _accessMode = ACCESS_Write;
		InstallCodec("cp932", true);
	}
	return true;
}

void File::Close()
{
	if (_needCloseFlag && _hFile != INVALID_HANDLE_VALUE) {
		::CloseHandle(_hFile);
		_hFile = INVALID_HANDLE_VALUE;
		_needCloseFlag = false;
		_fileName = "<invalid>";
		ReleaseCodec();
	}
}

int File::Seek(long offset, int origin)
{
	if (_hFile == INVALID_HANDLE_VALUE) return -1;
	DWORD dwMoveMethod =
		(origin == SEEK_SET)? FILE_BEGIN :
		(origin == SEEK_CUR)? FILE_CURRENT :
		(origin == SEEK_END)? FILE_END : FILE_BEGIN;
	return ::SetFilePointer(_hFile, offset, NULL, dwMoveMethod);
}

int File::GetRawChar()
{
	if (_hFile == INVALID_HANDLE_VALUE) return -1;
	DWORD dwBytesRead;
	unsigned char buff[1];
	::ReadFile(_hFile, buff, sizeof(buff), &dwBytesRead, NULL);
	return (dwBytesRead == 0)? -1 : buff[0];
}

void File::PutRawChar(char ch)
{
	if (_hFile == INVALID_HANDLE_VALUE) return;
	unsigned char buff[2];
	DWORD dwBytesToWrite, dwBytesWritten;
	dwBytesToWrite = 1;
	buff[0] = static_cast<unsigned char>(ch);
	::WriteFile(_hFile, buff, dwBytesToWrite, &dwBytesWritten, NULL);
}

size_t File::Read(void *buff, size_t bytes)
{
	DWORD dwBytesRead;
	::ReadFile(_hFile, buff, bytes, &dwBytesRead, NULL);
	return static_cast<size_t>(dwBytesRead);
}

size_t File::Write(const void *buff, size_t bytes)
{
	DWORD dwBytesWritten;
	::WriteFile(_hFile, buff, bytes, &dwBytesWritten, NULL);
	return static_cast<size_t>(dwBytesWritten);
}

bool File::GetFileStatByHandle(Signal sig, FileStat &fileStat) const
{
	unsigned long attr = 0;
	BY_HANDLE_FILE_INFORMATION attrData;
	String pathName = MakeAbsPath(_fileName.c_str(), NULL);
	if (::GetFileInformationByHandle(_hFile, &attrData) == 0) {
		sig.SetError(ERR_IOError, "failed to get file status of %s", pathName.c_str());
		return false;
	}
	fileStat = FileStat(pathName.c_str(), attrData);
	return true;
}

bool File::GetFileStat(Signal sig, const char *fileName, FileStat &fileStat)
{
	unsigned long attr = 0;
	WIN32_FILE_ATTRIBUTE_DATA attrData;
	String pathName = MakeAbsPath(fileName, NULL);
	if (::GetFileAttributesEx(pathName.c_str(), GetFileExInfoStandard, &attrData) == 0) {
		sig.SetError(ERR_IOError, "failed to get file status of %s", pathName.c_str());
		return false;
	}
	fileStat = FileStat(pathName.c_str(), attrData);
	return true;
}

bool File::IsExist(const char *pathName)
{
	WIN32_FILE_ATTRIBUTE_DATA attrData;
	return ::GetFileAttributesEx(pathName, GetFileExInfoStandard, &attrData) != 0;
}

#else // !defined(HAVE_WINDOWS_H)
const char File::Separator = SeparatorUnix;

File::File() : _fp(NULL), _needCloseFlag(false), _accessMode(ACCESS_Read)
{
}

bool File::Open(Signal sig, const char *fileName,
									const char *mode, const char *encoding)
{
	Close();
	bool processEOLFlag = false;
	do {
		const char *p = mode;
		if (*p == 'r') {
			_accessMode = ACCESS_Read;
		} else if (*p == 'w') {
			_accessMode = ACCESS_Write;
		} else if (*p == 'a') {
			_accessMode = ACCESS_Write;
		} else {
			sig.SetError(ERR_IOError, "invalid open mode");
			return false;
		}
		p++;
		for ( ; *p != '\0'; p++) {
			char ch = *p;
			if (ch == '+') {
				_accessMode = ACCESS_ReadWrite;
			} else if (ch == 't') {
				processEOLFlag = true;
			} else if (ch == 'b') {
				processEOLFlag = false;
			} else {
				sig.SetError(ERR_IOError, "invalid open mode");
				return false;
			}
		}
	} while (0);
	if (!InstallCodec(encoding, processEOLFlag)) {
		sig.SetError(ERR_CodecError, "unsupported encoding '%s'", encoding);
		return false;
	}
	_fileName = MakeAbsPath(fileName);
	_fp = ::fopen(_fileName.c_str(), mode);
	if (_fp == NULL) {
		sig.SetError(ERR_IOError,
				"can't open file '%s'", ExtractBaseName(fileName).c_str());
		return false;
	}
	_needCloseFlag = true;
	return true;
}

bool File::OpenStdin()
{
	_fp = stdin;
	_fileName = "stdin", _accessMode = ACCESS_Read;
	InstallCodec(Codec::EncodingFromLANG(), false);
	return true;
}

bool File::OpenStdout()
{
	_fp = stdout;
	_fileName = "stdout", _accessMode = ACCESS_Write;
	InstallCodec(Codec::EncodingFromLANG(), false);
	return true;
}

bool File::OpenStderr()
{
	_fp = stderr;
	_fileName = "stderr", _accessMode = ACCESS_Write;
	InstallCodec(Codec::EncodingFromLANG(), false);
	return true;
}

void File::Close()
{
	if (_needCloseFlag && _fp != NULL) {
		::fclose(_fp);
		_fp = NULL;
		_needCloseFlag = false;
		_fileName = "<invalid>";
		ReleaseCodec();
	}
}

int File::Seek(long offset, int origin)
{
	if (_fp == NULL) return -1;
	return ::fseek(_fp, offset, origin);
}

int File::GetRawChar()
{
	if (_fp == NULL) return -1;
	int ch = ::fgetc(_fp);
	return ch;
}

void File::PutRawChar(char ch)
{
	if (_fp == NULL) return;
	::fputc(ch, _fp);
}

size_t File::Read(void *buff, size_t bytes)
{
	return ::fread(buff, 1, bytes, _fp);
}

size_t File::Write(const void *buff, size_t bytes)
{
	return ::fwrite(buff, 1, bytes, _fp);
}

bool File::GetFileStatByHandle(Signal sig, FileStat &fileStat) const
{
	unsigned long attr = 0;
	struct stat stat;
	String pathName = MakeAbsPath(_fileName.c_str(), NULL);
	if (::fstat(fileno(_fp), &stat) != 0) {
		sig.SetError(ERR_IOError, "failed to get file status of %s", pathName.c_str());
		return false;
	}
	fileStat = FileStat(pathName.c_str(), stat);
	return true;
}

bool File::GetFileStat(Signal sig, const char *fileName, FileStat &fileStat)
{
	unsigned long attr = 0;
	struct stat stat;
	String pathName = MakeAbsPath(fileName, NULL);
	if (::stat(pathName.c_str(), &stat) != 0) {
		sig.SetError(ERR_IOError, "failed to get file status of %s", pathName.c_str());
		return false;
	}
	fileStat = FileStat(pathName.c_str(), stat);
	return true;
}

bool File::IsExist(const char *pathName)
{
	struct stat stat;
	return ::stat(pathName, &stat) == 0;
}

#endif

#if 0
void File::Print(const char *str)
{
	for ( ; *str != '\0'; str++) PutChar(*str);
}

void File::Println(const char *str)
{
	for ( ; *str != '\0'; str++) PutChar(*str);
	PutChar('\n');
}

void File::Printf(Signal sig, const char *format, const ValueList &valList)
{
	File::Formatter(*this).Format(sig, format, valList);
}
#endif

String File::ExtractDirName(const char *pathName)
{
	const char *p = pathName + ::strlen(pathName);
	for ( ; p >= pathName; p--) {
		if (IsSeparator(*p)) {
			String str;
			return RegulateName(str, pathName, p - pathName);
		}
	}
	return String("");
}

String File::ExtractBaseName(const char *pathName)
{
	const char *p = pathName + ::strlen(pathName);
	for ( ; p >= pathName; p--) {
		if (IsSeparator(*p)) {
			String str;
			return RegulateName(str, p + 1);
		}
	}
	return String(pathName);
}

const char *File::ExtractExtName(const char *pathName)
{
	const char *pBtm = pathName + ::strlen(pathName);
	for (const char *p = pBtm; p >= pathName; p--) {
		if (*p == '.') {
			return (p == pathName || IsSeparator(*(p - 1)))? pBtm : p + 1;
		}
		if (IsSeparator(*p)) break;
	}
	return pBtm;
}

void File::Split(const char *pathName, StringList &pathNameSepList)
{
	String field;
	if (IsSeparator(*pathName)) field.push_back(Separator);
	for (const char *p = pathName; *p != '\0'; p++) {
		if (IsSeparator(*p)) {
			if (!field.empty()) {
				pathNameSepList.push_back(field);
				field.clear();
			}
		} else {
			field.push_back(*p);
		}
	}
	if (!field.empty()) pathNameSepList.push_back(field);
}

bool File::HasWildCard(const char *pathName)
{
	for (const char *p = pathName; *p != '\0'; p++) {
		if (*p == '*' || *p == '?' || *p == '[' || *p == ']') {
			return true;
		}
	}
	return false;
}

// dirName must be specified as an absolute path.
String File::MakeAbsPath(const char *fileName, const char *dirName)
{
	String str;
	String dirNameWk;
	if (dirName == NULL || dirName[0] == '\0' || dirName[0] == '.' &&
					(dirName[1] == '\0' ||
					 dirName[1] == Separator && dirName[2] == '\0')) {
		dirNameWk = OAL::getcwd();
	} else {
		dirNameWk = dirName;
	}
	if (IsSeparator(fileName[0])) {
		for (const char *p = dirNameWk.c_str();
						*p != '\0' && !IsSeparator(*p); p++) {
			str += *p;
		}
	} else if (fileName[0] != '\0' && fileName[1] == ':') {
		// absolute path in Windows style.
	} else {
		int lenDirName = dirNameWk.size();
		if (lenDirName > 0 && IsSeparator(dirNameWk[lenDirName - 1])) {
			lenDirName--;
		}
		for (;;) {
			if (*fileName == '.' &&
					(IsSeparator(*(fileName + 1)) || *(fileName + 1) == '\0')) {
				fileName += IsSeparator(*(fileName + 2))? 2 : 1;
			} else if (*fileName == '.' && *(fileName + 1) == '.' &&
					(IsSeparator(*(fileName + 2)) || *(fileName + 2) == '\0')) {
				int len = lenDirName;
				if (len > 0) len--;
				for ( ; len > 0 && !IsSeparator(dirNameWk[len]); len--) ;
				if (len > 0) lenDirName = len;
				fileName += IsSeparator(*(fileName + 2))? 3 : 2;
			} else {
				break;
			}
		}
		for (const char *p = dirNameWk.c_str(); lenDirName > 0; p++, lenDirName--) {
			str += IsSeparator(*p)? Separator : *p;
		}
		str += Separator;
	}
	return RegulateName(str, fileName);
}

String &File::RegulateName(String &result, const char *name)
{
	for (const char *p = name; *p != '\0'; p++) {
		result += IsSeparator(*p)? Separator : *p;
	}
	return result;
}

String &File::RegulateName(String &result, const char *name, size_t cnt)
{
	for (const char *p = name; *p != '\0' && cnt > 0; p++, cnt--) {
		result += IsSeparator(*p)? Separator : *p;
	}
	return result;
}

#if 0
const String &File::GetWorkingDir()
{
	if (_pWorkingDir == NULL) {
		char *pathName;
		int maxLen = 256;
		for (int i = 0; i < 16; i++, maxLen += 128) {
			pathName = ::getcwd(NULL, maxLen);
			if (pathName != NULL) break;
		}
		if (pathName == NULL) ::exit(1);
		size_t len = ::strlen(pathName);
		if (len > 0 && pathName[len - 1] == Separator) pathName[len - 1] = '\0';
		_pWorkingDir = new String(pathName);
		::free(pathName);
	}
	return *_pWorkingDir;
}
#endif

#if defined(HAVE_WINDOWS_H)
const String &File::GetBaseDir()
{
	static String *pBaseDir = NULL;
	if (pBaseDir == NULL) {
		char pathName[512];
		::GetModuleFileName(NULL, pathName, NUMBEROF(pathName)); // Win32 API
		char *p = pathName + ::strlen(pathName);
		for ( ; p >= pathName; p--) {
			if (IsSeparator(*p)) {
				*p = '\0';
				break;
			}
		}
		pBaseDir = new String(pathName);
	}
	return *pBaseDir;
}

#else
const String &File::GetBaseDir()
{
	static String *pBaseDir = NULL;
	if (pBaseDir == NULL) {
		char *dirName = ::getcwd(NULL, PATH_MAX);
		pBaseDir = new String(dirName);
		::free(dirName);
	}
	return *pBaseDir;
}
#endif

bool File::IsMatchName(const char *pattern, const char *fileName, bool ignoreCaseFlag)
{
	if (*pattern == '\0') {
		return *fileName == '\0';
	} else if (*pattern == '*') {
		return IsMatchName(pattern + 1, fileName, ignoreCaseFlag) ||
			*fileName != '\0' && IsMatchName(pattern, fileName + 1, ignoreCaseFlag);
	} else if (*pattern == '?') {
		return *fileName != '\0' && IsMatchName(pattern + 1, fileName + 1, ignoreCaseFlag);
	} else if (*pattern == '[') {
		bool negFlag = false;
		pattern++;
		if (*pattern == '!') {
			negFlag = true;
			pattern++;
			for ( ; *pattern != ']' && *pattern != '\0'; pattern++) {
				if (CompareChar(*fileName, *pattern, ignoreCaseFlag) == 0) return false;
			}
		} else {
			for ( ; *pattern != ']' && *pattern != '\0'; pattern++) {
				if (CompareChar(*fileName, *pattern, ignoreCaseFlag) != 0) return false;
			}
		}
		if (*pattern == ']') pattern++;
		return IsMatchName(pattern, fileName + 1, ignoreCaseFlag);
	} else {
		return CompareChar(*pattern, *fileName, ignoreCaseFlag) == 0 &&
						IsMatchName(pattern + 1, fileName + 1, ignoreCaseFlag);
	}
}

}
