//
// Object_List
//

#include "Object_List.h"
#include "Object_Function.h"
#include "Object_Bytes.h"
#include "Expr.h"
#include "combination.hpp"

//-----------------------------------------------------------------------------
// extra
//-----------------------------------------------------------------------------
namespace std {

// extracted from combination.hpp which is supposed to be used with boost
template <class BidirectionalIterator>
  bool next_partial_permutation(BidirectionalIterator first,
                                BidirectionalIterator middle,
                                BidirectionalIterator last)
{
  reverse (middle, last);
  return next_permutation(first, last);
}

}

namespace AScript {

//-----------------------------------------------------------------------------
// Object_List
//-----------------------------------------------------------------------------
bool Object_List::IsList() const { return true; }

Object *Object_List::Clone() const
{
	return new Object_List(*this);
}

Value Object_List::GetByIndex(Signal sig, const Value &valueIdx)
{
	if (!valueIdx.IsNumber()) {
		sig.SetError(ERR_IndexError, "index must be a number for list");
		return Value::Null;
	}
	size_t idx = valueIdx.GetSizeT();
	if (idx >= GetList().size()) {
		sig.SetError(ERR_IndexError, "index is out of range");
		return Value::Null;
	}
	return GetList()[idx];
}

void Object_List::SetByIndex(Signal sig, const Value &valueIdx, const Value &value)
{
	if (!valueIdx.IsNumber()) {
		sig.SetError(ERR_IndexError, "index must be a number for list");
		return;
	}
	size_t idx = valueIdx.GetSizeT();
	if (idx >= GetList().size()) {
		sig.SetError(ERR_IndexError, "index is out of range");
		return;
	}
	GetList()[idx] = value;
}

Iterator *Object_List::CreateIterator(Signal sig)
{
	return new IteratorEach(dynamic_cast<Object_List *>(IncRef()));
}

String Object_List::ToString(Signal sig, bool exprFlag)
{
	String str;
	const ValueList &valList = GetList();
	str += "[";
	foreach_const (ValueList, pValue, valList) {
		if (pValue != valList.begin()) str += ", ";
		str += pValue->ToString(sig);
	}
	str += "]";
	return str;
}

void Object_List::DoPadding(Signal sig, ValueList &valList, const Value &value)
{
	if (sig.IsSignalled()) return;	// detect stack overflow
	int sizeMax = -1;
	foreach (ValueList, pValue, valList) {
		if (pValue->IsList()) {
			int size = static_cast<int>(pValue->GetList().size());
			if (sizeMax < size) sizeMax = size;
		}
	}
	if (sizeMax < 0) return;
	foreach (ValueList, pValue, valList) {
		if (pValue->IsList()) {
			ValueList &valListElem = pValue->GetList();
			int sizeToPadding = sizeMax - static_cast<int>(valListElem.size());
			while (sizeToPadding-- > 0) {
				valListElem.push_back(value);
			}
			DoPadding(sig, valListElem, value);
		}
	}
}

Object_List *Object_List::SortRank(Signal sig, const Value &valDirective,
					const ValueList *pValListKey, bool rankFlag, bool stableFlag)
{
	typedef std::map<const Value *, const Value *> ValuePtrMap;
	Environment &env = *this;
	enum { MODE_Ascend, MODE_Descend, MODE_Custom } mode = MODE_Ascend;
	const Function *pFunc = NULL;
	const ValueList &valList = GetList();
	if (valDirective.IsInvalid()) {
		// nothing to do
	} else if (valDirective.IsSymbol()) {
		const Symbol *pSymbol = valDirective.GetSymbol();
		if (pSymbol->IsIdentical(AScript_Symbol(ascend))) {
			mode = MODE_Ascend;
		} else if (pSymbol->IsIdentical(AScript_Symbol(descend))) {
			mode = MODE_Descend;
		} else {
			sig.SetError(ERR_ValueError,
				"invalid symbol '%s'", pSymbol->GetName());
			return NULL;
		}
	} else if (valDirective.IsFunction()) {
		mode = MODE_Custom;
		pFunc = valDirective.GetFunction();
		if (pFunc->GetDeclList().size() != 2) {
			sig.SetError(ERR_TypeError,
				"only a binary function can be specified");
			return NULL;
		}
	} else {
		sig.SetError(ERR_TypeError, "invalid argument");
		return NULL;
	}
	ValuePtrList valPtrList;
	ValuePtrMap valuePtrMap;
	if (rankFlag || pValListKey == NULL) {
		foreach_const (ValueList, pValue, valList) {
			valPtrList.push_back(&(*pValue));
		}
	} else {
		ValueList::const_iterator pValueKey = pValListKey->begin();
		ValueList::const_iterator pValue = valList.begin();
		for ( ; pValueKey != pValListKey->end() && pValue != valList.end();
														pValueKey++, pValue++) {
			valPtrList.push_back(&(*pValueKey));
			valuePtrMap[&(*pValueKey)] = &(*pValue);
		}
	}
	if (stableFlag) {
		if (mode == MODE_Ascend) {
			std::stable_sort(valPtrList.begin(), valPtrList.end(), Comparator_Ascend());
		} else if (mode == MODE_Descend) {
			std::stable_sort(valPtrList.begin(), valPtrList.end(), Comparator_Descend());
		} else { // mode == MODE_Custom
			std::stable_sort(valPtrList.begin(), valPtrList.end(),
											Comparator_Custom(env, sig, pFunc));
		}
	} else {
		if (mode == MODE_Ascend) {
			std::sort(valPtrList.begin(), valPtrList.end(), Comparator_Ascend());
		} else if (mode == MODE_Descend) {
			std::sort(valPtrList.begin(), valPtrList.end(), Comparator_Descend());
		} else { // mode == MODE_Custom
			std::sort(valPtrList.begin(), valPtrList.end(),
											Comparator_Custom(env, sig, pFunc));
		}
	}
	Value result;
	ValueList &valListResult = result.InitAsList(env);
	if (rankFlag) {
		foreach_const (ValueList, pValue, valList) {
			ValuePtrList::iterator ppValue = valPtrList.begin();
			for ( ; ppValue != valPtrList.end(); ppValue++) {
				if (Value::Compare(*pValue, **ppValue) == 0) break;
			}
			if (ppValue == valPtrList.end()) {
				sig.SetError(ERR_SystemError, "fatal error in rank() operation");
				return NULL;
			} else {
				size_t idx = ppValue - valPtrList.begin();
				valListResult.push_back(Value(idx));
			}
		}
	} else if (pValListKey == NULL) {
		foreach_const (ValuePtrList, ppValue, valPtrList) {
			valListResult.push_back(**ppValue);
		}
	} else {
		foreach_const (ValuePtrList, ppValueKey, valPtrList) {
			const Value *pValueKey = *ppValueKey;
			valListResult.push_back(*valuePtrMap[pValueKey]);
		}
	}
	return dynamic_cast<Object_List *>(result.GetListObj()->IncRef());
}

void Object_List::ValueVisitorEx::Visit(Signal sig, const Value &value)
{
	ASSUME(_env, value.IsNumber());
	size_t idx = value.GetSizeT();
	if (std::find(_indexList.begin(), _indexList.end(), idx) != _indexList.end()) {
		// nothing to do
	} else if (idx < _valList.size()) {
		_indexList.push_back(idx);
	} else {
		sig.SetError(ERR_IndexError, "index is out of range");
	}
}

//-----------------------------------------------------------------------------
// Object_List::IteratorEach
//-----------------------------------------------------------------------------
Object_List::IteratorEach::~IteratorEach()
{
	Object::Delete(_pObj);
}

bool Object_List::IteratorEach::DoNext(Signal sig, Value &value)
{
	ValueList &valList = _pObj->GetList();
	if (_pValue == _pValueEnd) return false;
	value = *_pValue;
	_pValue++;
	return true;
}

String Object_List::IteratorEach::ToString(Signal sig) const
{
	return String("<iterator:list#each>");
}

//-----------------------------------------------------------------------------
// Object_List::IteratorReverse
//-----------------------------------------------------------------------------
Object_List::IteratorReverse::~IteratorReverse()
{
	Object::Delete(_pObj);
}

bool Object_List::IteratorReverse::DoNext(Signal sig, Value &value)
{
	ValueList &valList = _pObj->GetList();
	if (_pValue == valList.rend()) return false;
	value = *_pValue;
	_pValue++;
	return true;
}

String Object_List::IteratorReverse::ToString(Signal sig) const
{
	return String("<iterator:list#reverse>");
}

//-----------------------------------------------------------------------------
// Object_List::IteratorRound
//-----------------------------------------------------------------------------
Object_List::IteratorRound::~IteratorRound()
{
	Object::Delete(_pObj);
}

bool Object_List::IteratorRound::DoNext(Signal sig, Value &value)
{
	ValueList &valList = _pObj->GetList();
	if (_pValue == valList.end() || _cnt == 0) return false;
	value = *_pValue;
	_pValue++;
	if (_cnt > 0) _cnt--;
	if (_pValue == valList.end()) _pValue = valList.begin();
	return true;
}

String Object_List::IteratorRound::ToString(Signal sig) const
{
	return String("<iterator:list#round>");
}

//-----------------------------------------------------------------------------
// Object_List::IteratorFold
//-----------------------------------------------------------------------------
Object_List::IteratorFold::~IteratorFold()
{
	Object::Delete(_pObj);
}

bool Object_List::IteratorFold::DoNext(Signal sig, Value &value)
{
	ValueList &valList = _pObj->GetList();
	if (_offset >= valList.size()) return false;
	Environment &env = *_pObj;
	value.InitAsIterator(env, new IteratorEach(
			dynamic_cast<Object_List *>(_pObj->IncRef()), _offset, _cntPerFold));
	_offset += _cntStep;
	return true;
}

String Object_List::IteratorFold::ToString(Signal sig) const
{
	return String("<iterator:list#fold>");
}

//-----------------------------------------------------------------------------
// Object_List::IteratorPermutation
//-----------------------------------------------------------------------------
Object_List::IteratorPermutation::IteratorPermutation(Object_List *pObj, int cnt) :
							Iterator(false), _pObj(pObj), _cnt(cnt), _validFlag(true)
{
	ValueList &valList = _pObj->GetList();
	_indexList.reserve(valList.size());
	for (size_t index = 0; index < valList.size(); index++) {
		_indexList.push_back(index);
	}
}

Object_List::IteratorPermutation::~IteratorPermutation()
{
	Object::Delete(_pObj);
}

bool Object_List::IteratorPermutation::DoNext(Signal sig, Value &value)
{
	if (!_validFlag) return false;
	ValueList &valList = value.InitAsList(*_pObj);
	ValueList &valListSrc = _pObj->GetList();
	if (_cnt < 0) {
		foreach (IndexList, pIndex, _indexList) {
			valList.push_back(valListSrc[*pIndex]);
		}
		_validFlag = std::next_permutation(_indexList.begin(), _indexList.end());
	} else {
		IndexList::iterator pIndex = _indexList.begin();
		for (int i = 0; i < _cnt; pIndex++, i++) {
			valList.push_back(valListSrc[*pIndex]);
		}
		_validFlag = std::next_partial_permutation(
				_indexList.begin(), _indexList.begin() + _cnt, _indexList.end());
	}
	return true;
}

String Object_List::IteratorPermutation::ToString(Signal sig) const
{
	return String("<iterator:list#permutation>");
}

//-----------------------------------------------------------------------------
// Object_List::IteratorCombination
//-----------------------------------------------------------------------------
Object_List::IteratorCombination::IteratorCombination(Object_List *pObj, int cnt) :
							Iterator(false), _pObj(pObj), _cnt(cnt), _validFlag(true)
{
	ValueList &valList = _pObj->GetList();
	_indexList.reserve(valList.size());
	for (size_t index = 0; index < valList.size(); index++) {
		_indexList.push_back(index);
	}
}

Object_List::IteratorCombination::~IteratorCombination()
{
	Object::Delete(_pObj);
}

bool Object_List::IteratorCombination::DoNext(Signal sig, Value &value)
{
	if (!_validFlag) return false;
	ValueList &valList = value.InitAsList(*_pObj);
	ValueList &valListSrc = _pObj->GetList();
	IndexList::iterator pIndex = _indexList.begin();
	for (int i = 0; i < _cnt; pIndex++, i++) {
		valList.push_back(valListSrc[*pIndex]);
	}
	// the following function is implemented in combination.hpp
	_validFlag = boost::next_combination(
			_indexList.begin(), _indexList.begin() + _cnt, _indexList.end());
	return true;
}

String Object_List::IteratorCombination::ToString(Signal sig) const
{
	return String("<iterator:list#combination>");
}

//-----------------------------------------------------------------------------
// Object_List::Comparator_Custom
//-----------------------------------------------------------------------------
bool Object_List::Comparator_Custom::
				operator()(const Value *pValue1, const Value *pValue2) const
{
	if (_sig.IsSignalled()) return false;
	ValueList valListArg(*pValue1, *pValue2);
	Context contextSub(valListArg);
	Value value = _pFunc->Eval(_env, _sig, contextSub);
	return value.GetNumber() < 0;
}

//-----------------------------------------------------------------------------
// AScript interfaces for Object_List
//-----------------------------------------------------------------------------
// list#clear!()
AScript_DeclareMethodEx(List, clear_X, "clear!")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	SetHelp("Clear the content of the list.");
}

AScript_ImplementMethod(List, clear_X)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	pSelf->GetList().clear();
	return context.GetSelf();
}

// list#shuffle!()
AScript_DeclareMethodEx(List, shuffle_X, "shuffle!")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	SetHelp("Shuffle the order of the list content based on random numbers.");
}

AScript_ImplementMethod(List, shuffle_X)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	ValueList &valList = pSelf->GetList();
	RandomNumberGenerator randomNumberGenerator;
	std::random_shuffle(valList.begin(), valList.end(), randomNumberGenerator);
	return context.GetSelf();
}

// list#add!(elem+)
AScript_DeclareMethodEx(List, add_X, "add!")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "elem", VTYPE_Any, OCCUR_OnceOrMore);
	SetHelp("Add specified items to the list.");
}

AScript_ImplementMethod(List, add_X)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	ValueList &valList = pSelf->GetList();
	foreach_const (ValueList, pValue, context.GetList(0)) {
		valList.push_back(*pValue);
	}
	return context.GetSelf();
}

// list#append!(elem+)
AScript_DeclareMethodEx(List, append_X, "append!")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "elem", VTYPE_Any, OCCUR_OnceOrMore);
	SetHelp(
	"Adds specified items to the list. If the item is a list or an iterator,\n"
	"each element in such an item is added to the list.");
}

AScript_ImplementMethod(List, append_X)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	ValueList &valList = pSelf->GetList();
	foreach_const (ValueList, pValue, context.GetList(0)) {
		if (pValue->IsList() || pValue->IsIterator()) {
			Iterator *pIterator = pValue->CreateIterator(sig);
			if (sig.IsSignalled()) return Value::Null;
			Value value;
			while (pIterator->Next(sig, value)) {
				valList.push_back(value);
			}
			Iterator::Delete(pIterator);
			if (sig.IsSignalled()) return Value::Null;
		} else {
			valList.push_back(*pValue);
		}
	}
	return context.GetSelf();
}

// list#shift!():[raise]
AScript_DeclareMethodEx(List, shift_X, "shift!")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareAttr(AScript_Symbol(raise));
	SetHelp(
	"Shifts the elements of the list. If the content of the list is [1, 2, 3, 4],\n"
	"it becomes [2, 3, 4] after calling this method. In default, no error occurs\n"
	"even when the list is empty. To raise an error for executing this method on\n"
	"an empty list, specify :raise attribute.");
}

AScript_ImplementMethod(List, shift_X)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	ValueList &valList = pSelf->GetList();
	if (valList.empty()) {
		if (context.IsSet(AScript_Symbol(raise))) {
			sig.SetError(ERR_ValueError, "no items");
		}
		return Value::Null;
	}
	Value result = valList.front();
	valList.erase(valList.begin());
	return result;
}

// list#erase!(idx+:number)
AScript_DeclareMethodEx(List, erase_X, "erase!")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "idx", VTYPE_Number, OCCUR_OnceOrMore);
	SetHelp("Erases elements at the specified indices.");
}

AScript_ImplementMethod(List, erase_X)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	ValueList &valList = pSelf->GetList();
	Object_List::ValueVisitorEx visitor(env, valList);
	foreach_const (ValueList, pValue, context.GetList(0)) {
		pValue->Accept(sig, visitor);
	}
	Object_List::IndexList &indexList = visitor.GetIndexList();
	if (!indexList.empty()) {
		std::sort(indexList.begin(), indexList.end());
		size_t offset = 0;
		foreach_const (Object_List::IndexList, pIdx, indexList) {
			size_t idx = *pIdx;
			valList.erase(valList.begin() + idx - offset);
			offset++;
		}
	}
	return context.GetSelf();
}

// list#padding!(value)
AScript_DeclareMethodEx(List, padding_X, "padding!")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "value", VTYPE_Any);
}

AScript_ImplementMethod(List, padding_X)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List::DoPadding(sig, pSelf->GetList(), context.GetValue(0));
	if (sig.IsSignalled()) return Value::Null;
	return context.GetSelf();
}

// list#get(index):map
AScript_DeclareMethod(List, get)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "index", VTYPE_Any);
}

AScript_ImplementMethod(List, get)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	return pSelf->GetByIndex(sig, context.GetValue(0));
}

// list#first()
AScript_DeclareMethod(List, first)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
}

AScript_ImplementMethod(List, first)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	ValueList &valList = pSelf->GetList();
	if (valList.empty()) {
		sig.SetError(ERR_ValueError, "list is empty");
		return Value::Null;
	}
	return valList.front();
}

// list#last()
AScript_DeclareMethod(List, last)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
}

AScript_ImplementMethod(List, last)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	ValueList &valList = pSelf->GetList();
	if (valList.empty()) {
		sig.SetError(ERR_ValueError, "list is empty");
		return Value::Null;
	}
	return valList.back();
}

// list#permutation(n?:number) {block?}
AScript_DeclareMethod(List, permutation)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "n", VTYPE_Number, OCCUR_ZeroOrOnce);
	DeclareBlock(OCCUR_ZeroOrOnce);
	SetHelp(
	"Creates an iterator that generates lists that contain elements picked up\n"
	"from the original list in a permutation manner.\n"
	ITERATOR_HELP
	"Block parameter format: |value:list, idx:number|");
}

AScript_ImplementMethod(List, permutation)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	int cnt = context.IsNumber(0)? context.GetInt(0) : -1;
	if (cnt > 0 && pSelf->GetList().size() < static_cast<size_t>(cnt)) {
		sig.SetError(ERR_ValueError, "specified size is out of range");
		return Value::Null;
	}
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorPermutation(pObj, cnt);
	return ReturnIterator(env, sig, context, pIterator);
}

// list#combination(n:number) {block?}
AScript_DeclareMethod(List, combination)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "n", VTYPE_Number);
	DeclareBlock(OCCUR_ZeroOrOnce);
	SetHelp(
	"Creates an iterator that generates lists that contain elements picked up\n"
	"from the original list in a combination manner.\n"
	ITERATOR_HELP
	"Block parameter format: |value:list, idx:number|");
}

AScript_ImplementMethod(List, combination)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	int cnt = context.GetInt(0);
	if (pSelf->GetList().size() < static_cast<size_t>(cnt)) {
		sig.SetError(ERR_ValueError, "specified size is out of range");
		return Value::Null;
	}
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorCombination(pObj, cnt);
	return ReturnIterator(env, sig, context, pIterator);
}

// list#len()
AScript_DeclareMethod(List, len)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	SetHelp("Returns the length of the list.");
}

AScript_ImplementMethod(List, len)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	size_t cnt = pSelf->GetList().size();
	return Value(cnt);
}

// list#min():[index,last_index,indices]
AScript_DeclareMethod(List, min)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareAttr(AScript_Symbol(index));
	DeclareAttr(AScript_Symbol(last_index));
	DeclareAttr(AScript_Symbol(indices));
	SetHelp(
	"Returns the minimum value in the list when no attribute is specified.\n"
	"With an attribute :index, it returns an index of the minimum value.\n"
	"With an attribute :last_index, it returns the last index of the minimum value\n"
	"when more than one elements have the same value.\n"
	"With an attribute :indices, it returns a list of indices of elements that\n"
	"has the minimum value.");
}

AScript_ImplementMethod(List, min)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIterator = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	Value result = pIterator->MinMax(env, sig, false, context.GetAttrs());
	Iterator::Delete(pIterator);
	if (sig.IsSignalled()) return Value::Null;
	return result;
}

// list#max():[index,last_index,indices]
AScript_DeclareMethod(List, max)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareAttr(AScript_Symbol(index));
	DeclareAttr(AScript_Symbol(last_index));
	DeclareAttr(AScript_Symbol(indices));
	SetHelp(
	"Returns the maximum value in the list when no attribute is specified.\n"
	"With an attribute :index, it returns an index of the maximum value.\n"
	"With an attribute :last_index, it returns the last index of the maximum value\n"
	"when more than one elements have the same value.\n"
	"With an attribute :indices, it returns a list of indices of elements that\n"
	"has the maximum value.");
}

AScript_ImplementMethod(List, max)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIterator = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	Value result = pIterator->MinMax(env, sig, true, context.GetAttrs());
	Iterator::Delete(pIterator);
	if (sig.IsSignalled()) return Value::Null;
	return result;
}

// list#sum()
AScript_DeclareMethod(List, sum)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	SetHelp("Returns a sum of values in the list.");
}

AScript_ImplementMethod(List, sum)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIterator = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	size_t cnt;
	Value result = pIterator->Sum(sig, cnt);
	Iterator::Delete(pIterator);
	if (sig.IsSignalled()) return Value::Null;
	return result;
}

// list#average()
AScript_DeclareMethod(List, average)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	SetHelp("Returns an average of values in the list.");
}

AScript_ImplementMethod(List, average)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIterator = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	size_t cnt;
	Value result = pIterator->Average(sig, cnt);
	Iterator::Delete(pIterator);
	if (sig.IsSignalled()) return Value::Null;
	return result;
}

// list#variance()
AScript_DeclareMethod(List, variance)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	SetHelp("Returns a variance of values in the list.");
}

AScript_ImplementMethod(List, variance)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIterator = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	size_t cnt;
	Value result = pIterator->Variance(sig, cnt);
	Iterator::Delete(pIterator);
	if (sig.IsSignalled()) return Value::Null;
	return result;
}

// list#stddev()
AScript_DeclareMethod(List, stddev)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	SetHelp("Returns a standard deviation of values in the list.");
}

AScript_ImplementMethod(List, stddev)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIterator = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	size_t cnt;
	Value result = pIterator->StandardDeviation(sig, cnt);
	Iterator::Delete(pIterator);
	if (sig.IsSignalled()) return Value::Null;
	return result;
}

// list#and()
AScript_DeclareMethodEx(List, and_, "and")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	SetHelp(
	"Calculates a logical AND result of all the values in the list.\n"
	"Values of boolean type's false and nil are recognized as false\n"
	"while others are true.");
}

AScript_ImplementMethod(List, and_)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIterator = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	Value result = pIterator->And(sig);
	Iterator::Delete(pIterator);
	if (sig.IsSignalled()) return Value::Null;
	return result;
}

// list#or()
AScript_DeclareMethodEx(List, or_, "or")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	SetHelp(
	"Calculates a logical OR result of all the values in the list.\n"
	"Values of boolean type's false and nil are recognized as false\n"
	"while others are true.");
}

AScript_ImplementMethod(List, or_)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIterator = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	Value result = pIterator->Or(sig);
	Iterator::Delete(pIterator);
	if (sig.IsSignalled()) return Value::Null;
	return result;
}

// list#filter(criteria) {block?}
AScript_DeclareMethod(List, filter)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "criteria", VTYPE_Any);
	DeclareBlock(OCCUR_ZeroOrOnce);
	SetHelp(
	"Returns a list that consists of elements of the original list after filtering\n"
	"by a criteria. A criteria can be an iterator or a function object.\n"
	"When an iterator is supplied as a criteria, it picks up true value in the iterator\n"
	"and creates a list that contains elements at corresponding position in the original list\n"
	"When a function object is applied, each element is passed to the function\n"
	"as an argument, and it collects values with evaluated results being true.");
}

AScript_ImplementMethod(List, filter)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIteratorSrc = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	Iterator *pIterator = pIteratorSrc->Filter(env, sig, context.GetValue(0));
	if (sig.IsSignalled()) {
		Iterator::Delete(pIteratorSrc);
		return Value::Null;
	}
	return ReturnIterator(env, sig, context, pIterator);
}

// list#while(criteria) {block?}
AScript_DeclareMethodEx(List, while_, "while")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "criteria", VTYPE_Any);
	DeclareBlock(OCCUR_ZeroOrOnce);
	SetHelp(
	"Creates an iterator that picks up each element in the list while criteria\n"
	"is evaluated as true. You can specify a function object, a list or an iterator\n"
	"as the criteria.\n"
	ITERATOR_HELP
	"Block parameter format: |value:list, idx:number|");
}

AScript_ImplementMethod(List, while_)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIteratorSrc = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	Iterator *pIterator = pIteratorSrc->While(env, sig, context.GetValue(0));
	if (sig.IsSignalled()) {
		Iterator::Delete(pIteratorSrc);
		return Value::Null;
	}
	return ReturnIterator(env, sig, context, pIterator);
}

// list#since(criteria) {block?}
AScript_DeclareMethod(List, since)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "criteria", VTYPE_Any);
	DeclareBlock(OCCUR_ZeroOrOnce);
	SetHelp(
	"Creates an iterator that picks up each element in the list since criteria\n"
	"is evaluated as true. You can specify a function object, a list or an iterator\n"
	"as the criteria.\n"
	ITERATOR_HELP
	"Block parameter format: |value:list, idx:number|");
}

AScript_ImplementMethod(List, since)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIteratorSrc = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	Iterator *pIterator = pIteratorSrc->Since(env, sig, context.GetValue(0));
	if (sig.IsSignalled()) {
		Iterator::Delete(pIteratorSrc);
		return Value::Null;
	}
	return ReturnIterator(env, sig, context, pIterator);
}

// list#map(func:Function) {block?}
AScript_DeclareMethod(List, map)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "func", VTYPE_Function);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(List, map)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIteratorSrc = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	Iterator *pIterator = new Iterator_ExplicitMap(env, pIteratorSrc,
			dynamic_cast<Object_Function *>(context.GetFunctionObj(0)->IncRef()));
	return ReturnIterator(env, sig, context, pIterator);
}

// list#reduce(accum) {block}
AScript_DeclareMethod(List, reduce)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "accum", VTYPE_Any);
	DeclareBlock(OCCUR_Once);
}

AScript_ImplementMethod(List, reduce)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIterator = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	Environment envBlock(&env, ENVTYPE_Block);
	const Function *pFuncBlock = GetBlockFunction(envBlock, sig, context);
	if (pFuncBlock == NULL) {
		Iterator::Delete(pIterator);
		return Value::Null;
	}
	Value result = pIterator->Reduce(env, sig, context.GetValue(0), pFuncBlock);
	Iterator::Delete(pIterator);
	if (sig.IsSignalled()) return Value::Null;
	return result;
}

// list#count(value)
AScript_DeclareMethod(List, count)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "value", VTYPE_Any);
}

AScript_ImplementMethod(List, count)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Iterator *pIterator = pSelf->CreateIterator(sig);
	if (sig.IsSignalled()) return Value::Null;
	size_t cnt = pIterator->Count(env, sig, context.GetValue(0));
	Iterator::Delete(pIterator);
	if (sig.IsSignalled()) return Value::Null;
	return Value(cnt);
}

// list#sort(directive?, keys[]?):[stable]
AScript_DeclareMethod(List, sort)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "directive", VTYPE_Any, OCCUR_ZeroOrOnce);
	DeclareArg(env, "keys", VTYPE_Any, OCCUR_ZeroOrOnce, true);
	DeclareAttr(AScript_Symbol(stable));
}

AScript_ImplementMethod(List, sort)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List *pObj = pSelf->SortRank(sig, context.GetValue(0),
						context.IsList(1)? &context.GetList(1) : NULL,
						false, context.IsSet(AScript_Symbol(stable)));
	if (sig.IsSignalled()) return Value::Null;
	Iterator *pIterator = new Object_List::IteratorEach(pObj);
	return ReturnIterator(env, sig, context, pIterator);
}

// list#rank(directive?):[stable]
AScript_DeclareMethod(List, rank)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "directive", VTYPE_Any, OCCUR_ZeroOrOnce);
	DeclareAttr(AScript_Symbol(stable));
}

AScript_ImplementMethod(List, rank)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List *pObj = pSelf->SortRank(sig, context.GetValue(0), NULL,
							true, context.IsSet(AScript_Symbol(stable)));
	if (sig.IsSignalled()) return Value::Null;
	Iterator *pIterator = new Object_List::IteratorEach(pObj);
	return ReturnIterator(env, sig, context, pIterator);
}

// list#join(sep:string => "")
AScript_DeclareMethod(List, join)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "sep", VTYPE_String, OCCUR_Once, false, false, new Expr_String(""));
}

AScript_ImplementMethod(List, join)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	ValueList &valList = pSelf->GetList();
	return Value(env, Join(valList, context.GetString(0)).c_str());
}

// list#format(format:string)
AScript_DeclareMethod(List, format)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "format", VTYPE_String);
}

AScript_ImplementMethod(List, format)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	return Value(env, Formatter::Format(sig,
						context.GetString(0), pSelf->GetList()).c_str());
}

// list#pack(format:string)
AScript_DeclareMethod(List, pack)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "format", VTYPE_String);
}

AScript_ImplementMethod(List, pack)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Value result;
	Object_Bytes *pObjBytes = result.InitAsBytes(env);
	size_t offset = 0;
	pObjBytes->Pack(sig, offset, context.GetString(0), pSelf->GetList());
	if (sig.IsSignalled()) return Value::Null;
	return result;
}

// list#each() {block?}
AScript_DeclareMethod(List, each)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareBlock(OCCUR_Once);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(List, each)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorEach(pObj);
	return ReturnIterator(env, sig, context, pIterator);
}

// list#offset(n:number):map {block?}
AScript_DeclareMethod(List, offset)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareBlock(OCCUR_Once);
	DeclareArg(env, "n", VTYPE_Number);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(List, offset)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	size_t offset = context.GetSizeT(0);
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorEach(pObj, offset);
	return ReturnIterator(env, sig, context, pIterator);
}

// list#align(n:number, value?):map {block?}
AScript_DeclareMethod(List, align)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "n", VTYPE_Number);
	DeclareArg(env, "value", VTYPE_Any, OCCUR_ZeroOrOnce);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(List, align)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorEach(pObj);
	pIterator = new Iterator_Align(pIterator, context.GetInt(0), context.GetValue(1));
	return ReturnIterator(env, sig, context, pIterator);
}

// list#skip(n:number):map {block?}
AScript_DeclareMethod(List, skip)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "n", VTYPE_Number);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(List, skip)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorEach(pObj);
	pIterator = new Iterator_Skip(pIterator, context.GetSizeT(0));
	return ReturnIterator(env, sig, context, pIterator);
}

// list#skipnil() {block?}
AScript_DeclareMethod(List, skipnil)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(List, skipnil)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorEach(pObj);
	pIterator = new Iterator_SkipInvalid(pIterator);
	return ReturnIterator(env, sig, context, pIterator);
}

// list#nilto(replace) {block?}
AScript_DeclareMethod(List, nilto)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "replace", VTYPE_Any);
}

AScript_ImplementMethod(List, nilto)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorEach(pObj);
	pIterator = new Iterator_ReplaceInvalid(pIterator, context.GetValue(0));
	return ReturnIterator(env, sig, context, pIterator);
}

// list#replace(value, replace) {block?}
AScript_DeclareMethod(List, replace)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "value", VTYPE_Any);
	DeclareArg(env, "replace", VTYPE_Any);
}

AScript_ImplementMethod(List, replace)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorEach(pObj);
	pIterator = new Iterator_Replace(pIterator,
								context.GetValue(0), context.GetValue(1));
	return ReturnIterator(env, sig, context, pIterator);
}

// list#head(n:number):map {block?}
AScript_DeclareMethod(List, head)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "n", VTYPE_Number);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(List, head)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	int cnt = context.GetInt(0);
	Iterator *pIterator = new Object_List::IteratorEach(pObj, 0, cnt);
	return ReturnIterator(env, sig, context, pIterator);
}

// list#tail(n:number):map {block?}
AScript_DeclareMethod(List, tail)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "n", VTYPE_Number);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(List, tail)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	int cnt = context.GetInt(0);
	int cntMax = static_cast<int>(pObj->GetList().size());
	size_t offset = (cntMax > cnt)? cntMax - cnt : cntMax;
	Iterator *pIterator = new Object_List::IteratorEach(pObj, offset);
	return ReturnIterator(env, sig, context, pIterator);
}

// list#reverse() {block?}
AScript_DeclareMethod(List, reverse)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(List, reverse)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorReverse(pObj);
	return ReturnIterator(env, sig, context, pIterator);
}

// list#round(n:number) {block?}
AScript_DeclareMethod(List, round)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "n", VTYPE_Number);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(List, round)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	int cnt = context.GetInt(0);
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorRound(pObj, cnt);
	return ReturnIterator(env, sig, context, pIterator);
}

// list#fold(n:number, nstep?:number) {block?}
AScript_DeclareMethod(List, fold)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "n", VTYPE_Number);
	DeclareArg(env, "nstep", VTYPE_Number, OCCUR_ZeroOrOnce);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(List, fold)
{
	Object_List *pSelf = Object_List::GetSelfObj(context);
	size_t cnt = context.GetSizeT(0);
	size_t cntStep = context.IsNumber(1)? context.GetSizeT(1) : cnt;
	Object_List *pObj = dynamic_cast<Object_List *>(pSelf->IncRef());
	Iterator *pIterator = new Object_List::IteratorFold(pObj, cnt, cntStep);
	return ReturnIterator(env, sig, context, pIterator);
}

// assignment
Class_List::Class_List(Environment &env) : Class(env.LookupClass(VTYPE_Object))
{
	AScript_AssignMethod(List, clear_X);
	AScript_AssignMethod(List, shuffle_X);
	AScript_AssignMethod(List, add_X);
	AScript_AssignMethod(List, append_X);
	AScript_AssignMethod(List, shift_X);
	AScript_AssignMethod(List, erase_X);
	AScript_AssignMethod(List, padding_X);
	AScript_AssignMethod(List, get);
	AScript_AssignMethod(List, first);
	AScript_AssignMethod(List, last);
	AScript_AssignMethod(List, permutation);
	AScript_AssignMethod(List, combination);
	// common operations with iterator
	AScript_AssignMethod(List, len);
	AScript_AssignMethod(List, min);
	AScript_AssignMethod(List, max);
	AScript_AssignMethod(List, sum);
	AScript_AssignMethod(List, average);
	AScript_AssignMethod(List, variance);
	AScript_AssignMethod(List, stddev);
	AScript_AssignMethod(List, and_);
	AScript_AssignMethod(List, or_);
	AScript_AssignMethod(List, filter);
	AScript_AssignMethod(List, while_);
	AScript_AssignMethod(List, since);
	AScript_AssignMethod(List, map);
	AScript_AssignMethod(List, reduce);
	AScript_AssignMethod(List, count);
	AScript_AssignMethod(List, sort);
	AScript_AssignMethod(List, rank);
	AScript_AssignMethod(List, join);
	AScript_AssignMethod(List, format);
	AScript_AssignMethod(List, pack);
	AScript_AssignMethod(List, each);
	AScript_AssignMethod(List, offset);
	AScript_AssignMethod(List, align);
	AScript_AssignMethod(List, skip);
	AScript_AssignMethod(List, skipnil);
	AScript_AssignMethod(List, nilto);
	AScript_AssignMethod(List, replace);
	AScript_AssignMethod(List, head);
	AScript_AssignMethod(List, tail);
	AScript_AssignMethod(List, reverse);
	AScript_AssignMethod(List, round);
	AScript_AssignMethod(List, fold);
}

Object *Class_List::CreateDescendant(Environment &env, Signal sig, Class *pClass)
{
	return new Object_List((pClass == NULL)? this : pClass);
}

}
